@@ -43,6 +43,8 @@ module System.Log.FastLogger (
   , module System.Log.FastLogger.Date
   -- * File rotation
   , module System.Log.FastLogger.File
+  -- * Types
+  , module System.Log.FastLogger.Types
   ) where
 
 #if __GLASGOW_HASKELL__ < 709
@@ -53,6 +55,7 @@ import Control.Concurrent (getNumCapabilities, myThreadId, threadCapability, tak
 import Control.Exception (handle, SomeException(..), bracket)
 import Control.Monad (when, replicateM)
 import Data.Array (Array, listArray, (!), bounds)
+import Data.Foldable (forM_)
 import Data.Maybe (isJust)
 import System.EasyFile (getFileSize)
 import System.Log.FastLogger.File
@@ -62,6 +65,7 @@ import System.Log.FastLogger.IORef
 import System.Log.FastLogger.LogStr
 import System.Log.FastLogger.Logger
 import System.Log.FastLogger.Date
+import System.Log.FastLogger.Types
 
 ----------------------------------------------------------------
 
@@ -175,6 +179,15 @@ rmLoggerSet (LoggerSet mfile fref arr _) = do
 type FastLogger = LogStr -> IO ()
 -- | 'TimedFastLogger' pass 'FormattedTime' to callback and simply log its result.
 -- this can be used to customize how to log timestamp.
+-- 
+-- Usually, one would write a wrapper on top of 'TimedFastLogger', for example:
+--
+-- @
+-- {-# LANGUAGE OverloadedStrings #-}
+--
+-- log :: TimedFastLogger -> LogStr -> IO ()
+-- log logger msg = logger (\time -> toLogStr (show time) <> " " <> msg <> "\n")
+-- @
 type TimedFastLogger = (FormattedTime -> LogStr) -> IO ()
 
 -- | Logger Type.
@@ -194,6 +207,11 @@ data LogType
                                   --   'BufSize' is a buffer size
                                   --   for each capability.
                                   --   File rotation is done on-demand.
+    | LogFileTimedRotate TimedFileLogSpec BufSize -- ^ Logging to a file.
+                                  --   'BufSize' is a buffer size
+                                  --   for each capability.
+                                  --   Rotation happens based on check specified
+                                  --   in 'TimedFileLogSpec'.
     | LogCallback (LogStr -> IO ()) (IO ()) -- ^ Logging with a log and flush action.
                                                -- run flush after log each message.
 
@@ -201,12 +219,13 @@ data LogType
 -- a tuple of logger and clean up action are returned.
 newFastLogger :: LogType -> IO (FastLogger, IO ())
 newFastLogger typ = case typ of
-    LogNone                  -> return (const noOp, noOp)
-    LogStdout bsize          -> newStdoutLoggerSet bsize >>= stdLoggerInit
-    LogStderr bsize          -> newStderrLoggerSet bsize >>= stdLoggerInit
-    LogFileNoRotate fp bsize -> newFileLoggerSet bsize fp >>= fileLoggerInit
-    LogFile fspec bsize      -> rotateLoggerInit fspec bsize
-    LogCallback cb flush     -> return (\str -> cb str >> flush, noOp)
+    LogNone                        -> return (const noOp, noOp)
+    LogStdout bsize                -> newStdoutLoggerSet bsize >>= stdLoggerInit
+    LogStderr bsize                -> newStderrLoggerSet bsize >>= stdLoggerInit
+    LogFileNoRotate fp bsize       -> newFileLoggerSet bsize fp >>= fileLoggerInit
+    LogFile fspec bsize            -> rotateLoggerInit fspec bsize
+    LogFileTimedRotate fspec bsize -> timedRotateLoggerInit fspec bsize
+    LogCallback cb flush           -> return (\str -> cb str >> flush, noOp)
   where
     stdLoggerInit lgrset = return (pushLogStr lgrset, rmLoggerSet lgrset)
     fileLoggerInit lgrset = return (pushLogStr lgrset, rmLoggerSet lgrset)
@@ -219,6 +238,18 @@ newFastLogger typ = case typ of
                 pushLogStr lgrset str
                 when (cnt <= 0) $ tryRotate lgrset fspec ref mvar
         return (logger, rmLoggerSet lgrset)
+    timedRotateLoggerInit fspec bsize = do
+        cache <- newTimeCache $ timed_timefmt fspec
+        now <- cache
+        lgrset <- newFileLoggerSet bsize $ prefixTime now $ timed_log_file fspec
+        ref <- newIORef now
+        mvar <- newMVar lgrset
+        let logger str = do
+                ct <- cache
+                updated <- updateTime (timed_same_timeframe fspec) ref ct
+                when updated $ tryTimedRotate fspec ct mvar
+                pushLogStr lgrset str
+        return (logger, rmLoggerSet lgrset)
 
 -- | 'bracket' version of 'newFastLogger'
 withFastLogger :: LogType -> (FastLogger -> IO a) -> IO a
@@ -231,12 +262,13 @@ newTimedFastLogger ::
                         -- "System.Log.FastLogger.Date" provide cached formatted time.
     -> LogType -> IO (TimedFastLogger, IO ())
 newTimedFastLogger tgetter typ = case typ of
-    LogNone                  -> return (const noOp, noOp)
-    LogStdout bsize          -> newStdoutLoggerSet bsize >>= stdLoggerInit
-    LogStderr bsize          -> newStderrLoggerSet bsize >>= stdLoggerInit
-    LogFileNoRotate fp bsize -> newFileLoggerSet bsize fp >>= fileLoggerInit
-    LogFile fspec bsize      -> rotateLoggerInit fspec bsize
-    LogCallback cb flush     -> return (\f -> tgetter >>= cb . f >> flush, noOp)
+    LogNone                        -> return (const noOp, noOp)
+    LogStdout bsize                -> newStdoutLoggerSet bsize >>= stdLoggerInit
+    LogStderr bsize                -> newStderrLoggerSet bsize >>= stdLoggerInit
+    LogFileNoRotate fp bsize       -> newFileLoggerSet bsize fp >>= fileLoggerInit
+    LogFile fspec bsize            -> rotateLoggerInit fspec bsize
+    LogFileTimedRotate fspec bsize -> timedRotateLoggerInit fspec bsize
+    LogCallback cb flush           -> return (\f -> tgetter >>= cb . f >> flush, noOp)
   where
     stdLoggerInit lgrset = return ( \f -> tgetter >>= pushLogStr lgrset . f, rmLoggerSet lgrset)
     fileLoggerInit lgrset = return (\f -> tgetter >>= pushLogStr lgrset . f, rmLoggerSet lgrset)
@@ -250,6 +282,19 @@ newTimedFastLogger tgetter typ = case typ of
                 pushLogStr lgrset (f t)
                 when (cnt <= 0) $ tryRotate lgrset fspec ref mvar
         return (logger, rmLoggerSet lgrset)
+    timedRotateLoggerInit fspec bsize = do
+        cache <- newTimeCache $ timed_timefmt fspec
+        now <- cache
+        lgrset <- newFileLoggerSet bsize $ prefixTime now $ timed_log_file fspec
+        ref <- newIORef now
+        mvar <- newMVar lgrset
+        let logger f = do
+                ct <- cache
+                updated <- updateTime (timed_same_timeframe fspec) ref ct
+                when updated $ tryTimedRotate fspec ct mvar
+                t <- tgetter
+                pushLogStr lgrset (f t)
+        return (logger, rmLoggerSet lgrset)
 
 -- | 'bracket' version of 'newTimeFastLogger'
 withTimedFastLogger :: IO FormattedTime -> LogType -> (TimedFastLogger -> IO a) -> IO a
@@ -263,6 +308,10 @@ noOp = return ()
 decrease :: IORef Int -> IO Int
 decrease ref = atomicModifyIORef' ref (\x -> (x - 1, x - 1))
 
+-- updateTime returns whether the timeframe has changed
+updateTime :: (FormattedTime -> FormattedTime -> Bool) -> IORef FormattedTime -> FormattedTime -> IO Bool
+updateTime cmp ref newTime = atomicModifyIORef' ref (\x -> (newTime, not $ cmp x newTime))
+
 tryRotate :: LoggerSet -> FileLogSpec -> IORef Int -> MVar () -> IO ()
 tryRotate lgrset spec ref mvar = bracket lock unlock rotateFiles
   where
@@ -292,3 +341,17 @@ tryRotate lgrset spec ref mvar = bracket lock unlock rotateFiles
         Just . fromIntegral <$> getFileSize file
     -- 200 is an ad-hoc value for the length of log line.
     estimate x = fromInteger (x `div` 200)
+
+
+tryTimedRotate :: TimedFileLogSpec -> FormattedTime -> MVar LoggerSet -> IO ()
+tryTimedRotate spec now mvar = bracket lock unlock rotateFiles
+  where
+    lock           = tryTakeMVar mvar
+    unlock Nothing = return ()
+    unlock (Just (LoggerSet current_path a b c)) = do
+        putMVar mvar $ LoggerSet (Just new_file_path) a b c
+        forM_ current_path (timed_post_process spec)
+    rotateFiles Nothing  = return ()
+    rotateFiles (Just (LoggerSet _ a b c)) = renewLoggerSet $ LoggerSet (Just new_file_path) a b c
+    new_file_path = prefixTime now $ timed_log_file spec
+
